@@ -20,6 +20,7 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <stdint.h>
 #include <glib.h>
 #include <glib-object.h>
 #include <glib/gprintf.h>
@@ -41,49 +42,6 @@ static gpointer jaw_impl_get_interface_data(JawObject *jaw_obj, guint iface);
 
 /* AtkObject */
 static void jaw_impl_initialize(AtkObject *atk_obj, gpointer data);
-static AtkObject* jaw_impl_get_parent(AtkObject *atk_obj);
-static AtkObject* jaw_impl_ref_child (AtkObject *atk_obj, gint i);
-static AtkRelationSet* jaw_impl_ref_relation_set(AtkObject *atk_obj);
-
-extern void jaw_action_interface_init(AtkActionIface*);
-extern gpointer jaw_action_data_init(jobject);
-extern void jaw_action_data_finalize(gpointer);
-
-extern void jaw_component_interface_init(AtkComponentIface*);
-extern gpointer jaw_component_data_init(jobject);
-extern void jaw_component_data_finalize(gpointer);
-
-extern void jaw_text_interface_init(AtkTextIface*);
-extern gpointer	jaw_text_data_init(jobject);
-extern void jaw_text_data_finalize(gpointer);
-
-extern void jaw_editable_text_interface_init(AtkEditableTextIface*);
-extern gpointer jaw_editable_text_data_init(jobject);
-extern void jaw_editable_text_data_finalize (gpointer);
-
-extern void jaw_hypertext_interface_init(AtkHypertextIface*);
-extern gpointer jaw_hypertext_data_init(jobject);
-extern void jaw_hypertext_data_finalize(gpointer);
-
-extern void jaw_image_interface_init(AtkImageIface*);
-extern gpointer jaw_image_data_init(jobject);
-extern void jaw_image_data_finalize(gpointer);
-
-extern void jaw_selection_interface_init(AtkSelectionIface*);
-extern gpointer jaw_selection_data_init(jobject);
-extern void jaw_selection_data_finalize(gpointer);
-
-extern void jaw_value_interface_init (AtkValueIface*);
-extern gpointer jaw_value_data_init (jobject);
-extern void jaw_value_data_finalize (gpointer);
-
-extern void jaw_table_interface_init (AtkTableIface*);
-extern gpointer jaw_table_data_init (jobject);
-extern void jaw_table_data_finalize (gpointer);
-
-extern void jaw_table_cell_interface_init (AtkTableCellIface*);
-extern gpointer jaw_table_cell_data_init (jobject);
-extern void jaw_table_cell_data_finalize (gpointer);
 
 typedef struct _JawInterfaceInfo {
   void (*finalize) (gpointer);
@@ -92,61 +50,120 @@ typedef struct _JawInterfaceInfo {
 
 static gpointer jaw_impl_parent_class = NULL;
 
+static GMutex typeTableMutex;
 static GHashTable *typeTable = NULL;
+static GMutex objectTableMutex;
 static GHashTable *objectTable = NULL;
-static gboolean jaw_debug = FALSE;
 
 static void
 object_table_insert (JNIEnv *jniEnv, jobject ac, JawImpl* jaw_impl)
 {
-  jclass classAccessibleContext = (*jniEnv)->FindClass( jniEnv,
-                                                       "javax/accessibility/AccessibleContext");
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "hashCode",
-                                          "()I");
-  gint hash_key = (gint)(*jniEnv)->CallIntMethod(jniEnv, ac, jmid);
-  g_hash_table_insert(objectTable, GINT_TO_POINTER(hash_key), GINT_TO_POINTER(jaw_impl));
+  JAW_DEBUG_C("%p, %p, %p", jniEnv, ac, jaw_impl);
+  jclass atkObject = (*jniEnv)->FindClass (jniEnv,"org/GNOME/Accessibility/AtkObject");
+  jmethodID jmid = (*jniEnv)->GetStaticMethodID (jniEnv, atkObject, "hashCode", "(Ljavax/accessibility/AccessibleContext;)I");
+  jaw_impl->hash_key = (gint)(*jniEnv)->CallStaticIntMethod (jniEnv, atkObject, jmid, ac);
+  g_mutex_lock(&objectTableMutex);
+  g_hash_table_insert(objectTable, GINT_TO_POINTER(jaw_impl->hash_key), jaw_impl);
+  g_mutex_unlock(&objectTableMutex);
 }
 
 static JawImpl*
 object_table_lookup (JNIEnv *jniEnv, jobject ac)
 {
-  jclass classAccessibleContext = (*jniEnv)->FindClass( jniEnv,
-                                                       "javax/accessibility/AccessibleContext" );
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "hashCode",
-                                          "()I" );
-  gint hash_key = (gint)(*jniEnv)->CallIntMethod( jniEnv, ac, jmid );
+  JAW_DEBUG_C("%p, %p", jniEnv, ac);
+  jclass atkObject = (*jniEnv)->FindClass (jniEnv,"org/GNOME/Accessibility/AtkObject");
+  jmethodID jmid = (*jniEnv)->GetStaticMethodID (jniEnv, atkObject, "hashCode", "(Ljavax/accessibility/AccessibleContext;)I");
+  gint hash_key = (gint)(*jniEnv)->CallStaticIntMethod (jniEnv, atkObject, jmid, ac);
   gpointer value = NULL;
+  g_mutex_lock(&objectTableMutex);
   if (objectTable==NULL)
+  {
+    g_mutex_unlock(&objectTableMutex);
     return NULL;
-
+  }
   value = g_hash_table_lookup(objectTable, GINT_TO_POINTER(hash_key));
+  g_mutex_unlock(&objectTableMutex);
   return (JawImpl*)value;
 }
 
 static void
-object_table_remove(JNIEnv *jniEnv, jobject ac)
+object_table_remove(JNIEnv *jniEnv, JawImpl *jaw_impl)
 {
-  jclass classAccessibleContext = (*jniEnv)->FindClass( jniEnv,
-                                                       "javax/accessibility/AccessibleContext" );
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "hashCode",
-                                          "()I" );
-  gint hash_key = (gint)(*jniEnv)->CallIntMethod( jniEnv, ac, jmid );
+  JAW_DEBUG_C("%p, %p", jniEnv, jaw_impl);
+  g_mutex_lock(&objectTableMutex);
+  g_hash_table_remove(objectTable, GINT_TO_POINTER(jaw_impl->hash_key));
+  g_mutex_unlock(&objectTableMutex);
+}
 
-  g_hash_table_remove(objectTable, GINT_TO_POINTER(hash_key));
+/* Called on completion of Java GC, take the opportunity to look for stale jaw_impl */
+void
+object_table_gc(JNIEnv *jniEnv)
+{
+  JAW_DEBUG_C("%p", jniEnv);
+  GHashTableIter iter;
+  gpointer key, value;
+  GSList *list = NULL, *cur, *next;
+
+  unsigned count[INTERFACE_MASK+1] = { 0, };
+
+  g_mutex_lock(&objectTableMutex);
+  if (objectTable)
+  {
+    g_hash_table_iter_init(&iter, objectTable);
+    while (g_hash_table_iter_next(&iter, &key, &value))
+    {
+      JawImpl *jaw_impl = value;
+      if ((*jniEnv)->IsSameObject(jniEnv, jaw_impl->parent.acc_context, NULL))
+      {
+	/* Got garbage-collected, mark for dropping */
+	list = g_slist_prepend(list, jaw_impl);
+      }
+      else
+      {
+	count[jaw_impl->tflag]++;
+      }
+    }
+  }
+  g_mutex_unlock(&objectTableMutex);
+
+  unsigned i;
+  for (i = 0; i < INTERFACE_MASK+1; i++) {
+    if (count[i] != 0) {
+      JAW_DEBUG_JNI("%x: %d", i, count[i]);
+    }
+  }
+
+  for (cur = list; cur != NULL; cur = next)
+  {
+    JawImpl *jaw_impl = cur->data;
+    g_object_unref(G_OBJECT(jaw_impl));
+    next = g_slist_next(cur);
+    g_slist_free_1(cur);
+  }
+}
+
+GHashTable*
+jaw_impl_get_object_hash_table(void)
+{
+  JAW_DEBUG_ALL("");
+  return objectTable;
+}
+
+GMutex*
+jaw_impl_get_object_hash_table_mutex(void)
+{
+  JAW_DEBUG_ALL("");
+  return &objectTableMutex;
 }
 
 static void
 aggregate_interface(JNIEnv *jniEnv, JawObject *jaw_obj, guint tflag)
 {
+  JAW_DEBUG_C("%p, %p, %u", jniEnv, jaw_obj, tflag);
   JawImpl *jaw_impl = JAW_IMPL(tflag, jaw_obj);
+  jaw_impl->tflag = tflag;
 
-  jobject ac = jaw_obj->acc_context;
+  jobject ac = (*jniEnv)->NewGlobalRef(jniEnv, jaw_obj->acc_context);
   jaw_impl->ifaceTable = g_hash_table_new(NULL, NULL);
 
   if (tflag & INTERFACE_ACTION)
@@ -248,28 +265,33 @@ aggregate_interface(JNIEnv *jniEnv, JawObject *jaw_obj, guint tflag)
                         (gpointer)INTERFACE_TABLE_CELL,
                         (gpointer)info);
   }
+
+  (*jniEnv)->DeleteGlobalRef(jniEnv, ac);
 }
 
 JawImpl*
 jaw_impl_get_instance (JNIEnv *jniEnv, jobject ac)
 {
+  JAW_DEBUG_C("%p, %p", jniEnv, ac);
   JawImpl *jaw_impl;
   jniEnv = jaw_util_get_jni_env();
 
   if (jniEnv == NULL)
     return NULL;
 
+  g_mutex_lock(&objectTableMutex);
   if (objectTable == NULL)
     objectTable = g_hash_table_new (NULL, NULL);
+  g_mutex_unlock(&objectTableMutex);
 
   jaw_impl = object_table_lookup(jniEnv, ac);
 
   if (jaw_impl == NULL)
   {
-    jobject global_ac = (*jniEnv)->NewGlobalRef(jniEnv, ac);
-    if (global_ac != NULL)
+    jobject temp_ref = (*jniEnv)->NewGlobalRef(jniEnv, ac);
+    if (temp_ref != NULL)
     {
-      guint tflag = jaw_util_get_tflag_from_jobj(jniEnv, global_ac);
+      guint tflag = jaw_util_get_tflag_from_jobj(jniEnv, temp_ref);
       jaw_impl = (JawImpl*)g_object_new(JAW_TYPE_IMPL(tflag), NULL);
       if (jaw_impl != NULL)
       {
@@ -277,42 +299,52 @@ jaw_impl_get_instance (JNIEnv *jniEnv, jobject ac)
 
         if (jaw_obj != NULL)
         {
-          jaw_obj->acc_context = global_ac;
+          jobject weak_ref = (*jniEnv)->NewWeakGlobalRef(jniEnv, temp_ref);
+          jaw_obj->acc_context = weak_ref;
           jaw_obj->storedData = g_hash_table_new(g_str_hash, g_str_equal);
           aggregate_interface(jniEnv, jaw_obj, tflag);
           atk_object_initialize(ATK_OBJECT(jaw_impl), NULL);
-          object_table_insert(jniEnv, global_ac, jaw_impl);
+          object_table_insert(jniEnv, weak_ref, jaw_impl);
         } else
         {
-          if (jaw_debug)
-            g_warning("jaw_impl_get_instance: jaw_obj == NULL");
+          JAW_DEBUG_I("jaw_obj == NULL");
+          (*jniEnv)->DeleteGlobalRef(jniEnv, temp_ref);
           return NULL;
         }
       } else
       {
-        if (jaw_debug)
-          g_warning("jaw_impl_get_instance: jaw_impl == NULL");
+        JAW_DEBUG_I("jaw_impl == NULL");
       }
+      (*jniEnv)->DeleteGlobalRef(jniEnv, temp_ref);
     } else
     {
-      if (jaw_debug)
-        g_warning("jaw_impl_get_instance: global_ac == NULL");
+      JAW_DEBUG_I("global_ac == NULL");
       return NULL;
     }
   }
   return jaw_impl;
 }
 
+JawImpl*
+jaw_impl_get_instance_from_jaw (JNIEnv *jniEnv, jobject ac)
+{
+  JAW_DEBUG_C("%p, %p", jniEnv, ac);
+  jclass classWrapper = (*jniEnv)->FindClass(jniEnv, "org/GNOME/Accessibility/AtkWrapper");
+  jmethodID jmid = (*jniEnv)->GetStaticMethodID(jniEnv, classWrapper, "getInstanceFromSwing", "(Ljavax/accessibility/AccessibleContext;)J");
+  jlong ptr = (*jniEnv)->CallStaticLongMethod(jniEnv, classWrapper, jmid, ac);
+  return (JawImpl*) (uintptr_t) ptr;
+}
+
 JawImpl*
 jaw_impl_find_instance (JNIEnv *jniEnv, jobject ac)
 {
+  JAW_DEBUG_C("%p, %p", jniEnv, ac);
   JawImpl *jaw_impl;
 
   jaw_impl = object_table_lookup(jniEnv, ac);
   if (jaw_impl == NULL)
   {
-    if (jaw_debug)
-      g_warning("jaw_impl_find_instance: jaw_impl");
+    JAW_DEBUG_I("jaw_impl == NULL");
     return NULL;
   }
 
@@ -320,8 +352,9 @@ jaw_impl_find_instance (JNIEnv *jniEnv, jobject ac)
 }
 
 static void
-jaw_impl_class_intern_init (gpointer klass)
+jaw_impl_class_intern_init (gpointer klass, gpointer data)
 {
+  JAW_DEBUG_ALL("%p, %p", klass, data);
   if (jaw_impl_parent_class == NULL)
   {
     jaw_impl_parent_class = g_type_class_peek_parent (klass);
@@ -333,6 +366,7 @@ jaw_impl_class_intern_init (gpointer klass)
 GType
 jaw_impl_get_type (guint tflag)
 {
+  JAW_DEBUG_C("%u", tflag);
   GType type;
 
   static const GInterfaceInfo atk_action_info =
@@ -356,7 +390,7 @@ jaw_impl_get_type (guint tflag)
     NULL
   };
 
-  static const GInterfaceInfo atk_editable_text_info = 
+  static const GInterfaceInfo atk_editable_text_info =
   {
     (GInterfaceInitFunc) jaw_editable_text_interface_init,
     (GInterfaceFinalizeFunc) NULL,
@@ -405,11 +439,13 @@ jaw_impl_get_type (guint tflag)
     NULL
   };
 
+  g_mutex_lock(&typeTableMutex);
   if (typeTable == NULL) {
     typeTable = g_hash_table_new( NULL, NULL );
   }
 
   type = GPOINTER_TO_GTYPE(g_hash_table_lookup(typeTable, GUINT_TO_POINTER(tflag)));
+  g_mutex_unlock(&typeTableMutex);
   if (type == 0) {
     GTypeInfo tinfo = {
       sizeof(JawImplClass),
@@ -459,7 +495,9 @@ jaw_impl_get_type (guint tflag)
     if (tflag & INTERFACE_TABLE_CELL)
       g_type_add_interface_static (type, ATK_TYPE_TABLE_CELL, &atk_table_cell_info);
 
+    g_mutex_lock(&typeTableMutex);
     g_hash_table_insert(typeTable, GINT_TO_POINTER(tflag), GTYPE_TO_POINTER(type));
+    g_mutex_unlock(&typeTableMutex);
   }
 
   return type;
@@ -468,29 +506,22 @@ jaw_impl_get_type (guint tflag)
 static void
 jaw_impl_class_init(JawImplClass *klass)
 {
+  JAW_DEBUG_ALL("%p", klass);
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   gobject_class->dispose = jaw_impl_dispose;
   gobject_class->finalize = jaw_impl_finalize;
 
   AtkObjectClass *atk_class = ATK_OBJECT_CLASS (klass);
   atk_class->initialize = jaw_impl_initialize;
-  atk_class->get_parent = jaw_impl_get_parent;
-  atk_class->ref_child = jaw_impl_ref_child;
-  atk_class->ref_relation_set = jaw_impl_ref_relation_set;
 
   JawObjectClass *jaw_class = JAW_OBJECT_CLASS (klass);
   jaw_class->get_interface_data = jaw_impl_get_interface_data;
 }
-/*
-static void
-jaw_impl_init(JawImpl *impl)
-{
-	jaw_impl->ifaceTable = g_hash_table_new(NULL, NULL);
-}
-*/
+
 static void
 jaw_impl_dispose(GObject *gobject)
 {
+  JAW_DEBUG_ALL("%p", gobject);
   /* Chain up to parent's dispose */
   G_OBJECT_CLASS(jaw_impl_parent_class)->dispose(gobject);
 }
@@ -498,31 +529,29 @@ jaw_impl_dispose(GObject *gobject)
 static void
 jaw_impl_finalize(GObject *gobject)
 {
+  JAW_DEBUG_ALL("%p", gobject);
   JawObject *jaw_obj = JAW_OBJECT(gobject);
-  jobject global_ac = jaw_obj->acc_context;
-
   JawImpl *jaw_impl = (JawImpl*)jaw_obj;
 
   JNIEnv *jniEnv = jaw_util_get_jni_env();
-  object_table_remove( jniEnv, global_ac );
+  object_table_remove( jniEnv, jaw_impl );
 
-  (*jniEnv)->DeleteGlobalRef(jniEnv, global_ac);
+  (*jniEnv)->DeleteWeakGlobalRef(jniEnv, jaw_obj->acc_context);
   jaw_obj->acc_context = NULL;
 
   /* Interface finalize */
-  GHashTableIter *iter = NULL;
-  gpointer *key = NULL;
-  gpointer *value = NULL;
+  GHashTableIter iter;
+  gpointer value;
 
-  g_hash_table_iter_init(iter, jaw_impl->ifaceTable);
-  while (g_hash_table_iter_next(iter, key, value))
+  g_hash_table_iter_init(&iter, jaw_impl->ifaceTable);
+  while (g_hash_table_iter_next(&iter, NULL, &value))
   {
     JawInterfaceInfo *info = (JawInterfaceInfo*)value;
     info->finalize(info->data);
 
     g_free(info);
 
-    g_hash_table_iter_remove(iter);
+    g_hash_table_iter_remove(&iter);
   }
   if (jaw_impl->ifaceTable != NULL)
   {
@@ -536,6 +565,7 @@ jaw_impl_finalize(GObject *gobject)
 static gpointer
 jaw_impl_get_interface_data (JawObject *jaw_obj, guint iface)
 {
+  JAW_DEBUG_C("%p, %u", jaw_obj, iface);
   JawImpl *jaw_impl = (JawImpl*)jaw_obj;
 
   if (jaw_impl->ifaceTable == NULL || jaw_impl == NULL)
@@ -552,11 +582,12 @@ jaw_impl_get_interface_data (JawObject *jaw_obj, guint iface)
 static void
 jaw_impl_initialize (AtkObject *atk_obj, gpointer data)
 {
+  JAW_DEBUG_C("%p, %p", atk_obj, data);
   ATK_OBJECT_CLASS(jaw_impl_parent_class)->initialize(atk_obj, data);
 
   JawObject *jaw_obj = JAW_OBJECT(atk_obj);
-  jobject ac = jaw_obj->acc_context;
   JNIEnv *jniEnv = jaw_util_get_jni_env();
+  jobject ac = (*jniEnv)->NewGlobalRef(jniEnv, jaw_obj->acc_context);
 
   jclass classAtkWrapper = (*jniEnv)->FindClass(jniEnv,
                                                 "org/GNOME/Accessibility/AtkWrapper");
@@ -565,244 +596,50 @@ jaw_impl_initialize (AtkObject *atk_obj, gpointer data)
                                                 "registerPropertyChangeListener",
                                                 "(Ljavax/accessibility/AccessibleContext;)V");
   (*jniEnv)->CallStaticVoidMethod(jniEnv, classAtkWrapper, jmid, ac);
-}
-
-static AtkObject*
-jaw_impl_get_parent (AtkObject *atk_obj)
-{
-  if (jaw_toplevel_get_child_index(JAW_TOPLEVEL(atk_get_root()), atk_obj) != -1)
-  {
-    return ATK_OBJECT(atk_get_root());
-  }
-
-  JawObject *jaw_obj = JAW_OBJECT(atk_obj);
-  jobject ac = jaw_obj->acc_context;
-  JNIEnv *jniEnv = jaw_util_get_jni_env();
-
-  jclass classAccessibleContext = (*jniEnv)->FindClass(jniEnv,
-                                                       "javax/accessibility/AccessibleContext" );
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "getAccessibleParent",
-                                          "()Ljavax/accessibility/Accessible;");
-  jobject jparent = (*jniEnv)->CallObjectMethod( jniEnv, ac, jmid );
-  if (jparent != NULL )
-  {
-    jclass classAccessible = (*jniEnv)->FindClass(jniEnv,
-                                                  "javax/accessibility/Accessible" );
-    jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                  classAccessible,
-                                  "getAccessibleContext",
-                                  "()Ljavax/accessibility/AccessibleContext;");
-    jobject parent_ac = (*jniEnv)->CallObjectMethod( jniEnv, jparent, jmid );
-
-    AtkObject *obj = (AtkObject*) object_table_lookup( jniEnv, parent_ac );
-    if (obj != NULL ) {
-       return obj;
-    }
-  }
-
-  return ATK_OBJECT(atk_get_root());
-}
-
-static AtkObject*
-jaw_impl_ref_child (AtkObject *atk_obj, gint i)
-{
-  JawObject *jaw_obj = JAW_OBJECT(atk_obj);
-  jobject ac = jaw_obj->acc_context;
-  JNIEnv *jniEnv = jaw_util_get_jni_env();
-
-  jclass classAccessibleContext = (*jniEnv)->FindClass(jniEnv,
-                                                       "javax/accessibility/AccessibleContext" );
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "getAccessibleChild",
-                                          "(I)Ljavax/accessibility/Accessible;" );
-  jobject jchild = (*jniEnv)->CallObjectMethod( jniEnv, ac, jmid, i );
-  if (jchild == NULL)
-  {
-    return NULL;
-  }
-
-  jclass classAccessible = (*jniEnv)->FindClass( jniEnv, "javax/accessibility/Accessible" );
-  jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                classAccessible,
-                                "getAccessibleContext",
-                                "()Ljavax/accessibility/AccessibleContext;" );
-  jobject child_ac = (*jniEnv)->CallObjectMethod( jniEnv, jchild, jmid );
-
-  AtkObject *obj = (AtkObject*) jaw_impl_get_instance( jniEnv, child_ac );
-  if (G_OBJECT(obj) != NULL)
-    g_object_ref(G_OBJECT(obj));
-
-  return obj;
-}
-
-static jstring
-get_java_relation_key_constant (JNIEnv *jniEnv, const gchar* strKey)
-{
-  jclass classAccessibleRelation = (*jniEnv)->FindClass(jniEnv,
-                                                        "javax/accessibility/AccessibleRelation");
-  jfieldID jfid = (*jniEnv)->GetStaticFieldID(jniEnv,
-                                              classAccessibleRelation,
-                                              strKey,
-                                              "Ljava/lang/String;");
-  jstring jkey = (*jniEnv)->GetStaticObjectField(jniEnv,
-                                                 classAccessibleRelation,
-                                                 jfid);
-
-  return jkey;
+  (*jniEnv)->DeleteGlobalRef(jniEnv, ac);
 }
 
 static gboolean
-is_java_relation_key (JNIEnv *jniEnv,jstring jKey, const gchar* strKey)
+is_java_relation_key (JNIEnv *jniEnv, jstring jKey, const gchar* strKey)
 {
-  jstring jConstKey = get_java_relation_key_constant (jniEnv, strKey);
-
-  if ( (*jniEnv)->IsSameObject(jniEnv, jKey, jConstKey) )
-  {
-    return TRUE;
-  } else
-  {
-    return FALSE;
-  }
+  JAW_DEBUG_C("%p, %p, %s", jniEnv, jKey, strKey);
+  jclass classAccessibleRelation = (*jniEnv)->FindClass(jniEnv, "javax/accessibility/AccessibleRelation");
+  jfieldID jfid = (*jniEnv)->GetStaticFieldID(jniEnv, classAccessibleRelation, strKey, "Ljava/lang/String;");
+  jstring jConstKey = (*jniEnv)->GetStaticObjectField(jniEnv, classAccessibleRelation, jfid);
+  return (*jniEnv)->IsSameObject(jniEnv, jKey, jConstKey);
 }
 
-static AtkRelationType
-get_atk_relation_type_from_java_key (JNIEnv *jniEnv, jstring jrel_key)
+AtkRelationType
+jaw_impl_get_atk_relation_type(JNIEnv *jniEnv, jstring jrel_key)
 {
+  JAW_DEBUG_C("%p, %p", jniEnv, jrel_key);
   if ( is_java_relation_key(jniEnv, jrel_key, "CHILD_NODE_OF") )
-  {
     return ATK_RELATION_NODE_CHILD_OF;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "CONTROLLED_BY") )
-  {
     return ATK_RELATION_CONTROLLED_BY;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "CONTROLLER_FOR") )
-  {
     return ATK_RELATION_CONTROLLER_FOR;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "EMBEDDED_BY") )
-  {
     return ATK_RELATION_EMBEDDED_BY;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "EMBEDS") )
-  {
     return ATK_RELATION_EMBEDS;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "FLOWS_FROM") )
-  {
     return ATK_RELATION_FLOWS_FROM;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "FLOWS_TO") )
-  {
     return ATK_RELATION_FLOWS_TO;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "LABEL_FOR") )
-  {
     return ATK_RELATION_LABEL_FOR;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "LABELED_BY") )
-  {
     return ATK_RELATION_LABELLED_BY;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "MEMBER_OF") )
-  {
     return ATK_RELATION_MEMBER_OF;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "PARENT_WINDOW_OF") )
-  {
     return ATK_RELATION_PARENT_WINDOW_OF;
-  }
   if ( is_java_relation_key(jniEnv, jrel_key, "SUBWINDOW_OF") )
-  {
     return ATK_RELATION_SUBWINDOW_OF;
-  }
-
   return ATK_RELATION_NULL;
 }
 
-static AtkRelationSet*
-jaw_impl_ref_relation_set (AtkObject *atk_obj)
-{
-  if (atk_obj->relation_set)
-    g_object_unref(G_OBJECT(atk_obj->relation_set));
-  atk_obj->relation_set = atk_relation_set_new();
-  if(atk_obj == NULL)
-    return NULL;
-
-  JawObject *jaw_obj = JAW_OBJECT(atk_obj);
-  jobject ac = jaw_obj->acc_context;
-  JNIEnv *jniEnv = jaw_util_get_jni_env();
-
-  jclass classAccessibleContext = (*jniEnv)->FindClass(jniEnv,
-                                                       "javax/accessibility/AccessibleContext" );
-  jmethodID jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                          classAccessibleContext,
-                                          "getAccessibleRelationSet",
-                                          "()Ljavax/accessibility/AccessibleRelationSet;" );
-  jobject jrel_set = (*jniEnv)->CallObjectMethod( jniEnv, ac, jmid );
-
-  jclass classAccessibleRelationSet = (*jniEnv)->FindClass( jniEnv,
-                                                           "javax/accessibility/AccessibleRelationSet");
-  jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                classAccessibleRelationSet,
-                                "toArray",
-                                "()[Ljavax/accessibility/AccessibleRelation;");
-  jobjectArray jrel_arr = (*jniEnv)->CallObjectMethod(jniEnv, jrel_set, jmid);
-  jsize jarr_size = (*jniEnv)->GetArrayLength(jniEnv, jrel_arr);
-
-  jsize i;
-  for (i = 0; i < jarr_size; i++)
-  {
-    jobject jrel = (*jniEnv)->GetObjectArrayElement(jniEnv, jrel_arr, i);
-    jclass classAccessibleRelation = (*jniEnv)->FindClass(jniEnv,
-                                                          "javax/accessibility/AccessibleRelation");
-    jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                  classAccessibleRelation,
-                                  "getKey",
-                                  "()Ljava/lang/String;");
-    jstring jrel_key = (*jniEnv)->CallObjectMethod( jniEnv, jrel, jmid );
-    AtkRelationType rel_type = get_atk_relation_type_from_java_key(jniEnv, jrel_key);
-
-    jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                  classAccessibleRelation,
-                                  "getTarget",
-                                  "()[Ljava/lang/Object;");
-    jobjectArray jtarget_arr = (*jniEnv)->CallObjectMethod(jniEnv, jrel, jmid);
-    jsize jtarget_size = (*jniEnv)->GetArrayLength(jniEnv, jtarget_arr);
-
-    jsize j;
-    for (j = 0; j < jtarget_size; j++)
-    {
-      jobject jtarget = (*jniEnv)->GetObjectArrayElement(jniEnv, jtarget_arr, j);
-      jclass classAccessible = (*jniEnv)->FindClass( jniEnv,
-                                                    "javax/accessibility/Accessible");
-      if ((*jniEnv)->IsInstanceOf(jniEnv, jtarget, classAccessible))
-      {
-        jmid = (*jniEnv)->GetMethodID(jniEnv,
-                                      classAccessible,
-                                      "getAccessibleContext",
-                                      "()Ljavax/accessibility/AccessibleContext;");
-        jobject target_ac = (*jniEnv)->CallObjectMethod(jniEnv, jtarget, jmid);
-
-        JawImpl *target_obj = jaw_impl_get_instance(jniEnv, target_ac);
-        if(target_obj == NULL)
-          return NULL;
-        atk_object_add_relationship(atk_obj, rel_type, (AtkObject*) target_obj);
-      }
-    }
-  }
-  if(atk_obj->relation_set == NULL)
-    return NULL;
-  if (G_OBJECT(atk_obj->relation_set) != NULL)
-    g_object_ref (atk_obj->relation_set);
-
-  return atk_obj->relation_set;
-}
-
 #ifdef __cplusplus
 }
 #endif
-
