@@ -22,9 +22,7 @@
 #include <string.h>
 #include <glib.h>
 #include "jawtoplevel.h"
-
-static void jaw_toplevel_class_init(JawToplevelClass *klass);
-static void jaw_toplevel_init(JawToplevel *toplevel);
+#include "jawutil.h"
 
 static void jaw_toplevel_initialize(AtkObject *accessible, gpointer data);
 static void jaw_toplevel_object_finalize(GObject *obj);
@@ -43,6 +41,7 @@ G_DEFINE_TYPE (JawToplevel, jaw_toplevel, ATK_TYPE_OBJECT)
 static void
 jaw_toplevel_class_init (JawToplevelClass *klass)
 {
+  JAW_DEBUG_ALL("%p", klass);
   AtkObjectClass *atk_object_class = ATK_OBJECT_CLASS(klass);
   GObjectClass *g_object_class = G_OBJECT_CLASS(klass);
 
@@ -61,18 +60,21 @@ jaw_toplevel_class_init (JawToplevelClass *klass)
 static void
 jaw_toplevel_init (JawToplevel *toplevel)
 {
+  JAW_DEBUG_ALL("%p", toplevel);
   toplevel->windows = NULL;
 }
 
 static void
 jaw_toplevel_initialize(AtkObject *accessible, gpointer data)
 {
+  JAW_DEBUG_ALL("%p, %p", accessible, data);
  ATK_OBJECT_CLASS (jaw_toplevel_parent_class)->initialize(accessible, data);
 }
 
 static void
 jaw_toplevel_object_finalize (GObject *obj)
 {
+  JAW_DEBUG_ALL("%p", obj);
   JawToplevel *jaw_toplevel = JAW_TOPLEVEL(obj);
   g_list_free(jaw_toplevel->windows);
 
@@ -83,6 +85,7 @@ jaw_toplevel_object_finalize (GObject *obj)
 static const gchar*
 jaw_toplevel_get_name (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   gint i;
   for(i = 0; i < atk_object_get_n_accessible_children(obj); i++)
   {
@@ -102,12 +105,14 @@ jaw_toplevel_get_name (AtkObject *obj)
 static const gchar*
 jaw_toplevel_get_description (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   return "Accessible Java application";
 }
 
 static gint
 jaw_toplevel_get_n_children (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   JawToplevel* jaw_toplevel = JAW_TOPLEVEL(obj);
   gint n = g_list_length(jaw_toplevel->windows);
 
@@ -117,6 +122,7 @@ jaw_toplevel_get_n_children (AtkObject *obj)
 static gint
 jaw_toplevel_get_index_in_parent (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   JawToplevel *jaw_toplevel = JAW_TOPLEVEL(obj);
   gint i = g_list_index(jaw_toplevel->windows, obj);
 
@@ -126,16 +132,19 @@ jaw_toplevel_get_index_in_parent (AtkObject *obj)
 static AtkRole
 jaw_toplevel_get_role (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   return ATK_ROLE_APPLICATION;
 }
 
 static AtkObject*
 jaw_toplevel_ref_child (AtkObject *obj, gint i)
 {
+  JAW_DEBUG_C("%p, %d", obj, i);
   JawToplevel *jaw_toplevel = JAW_TOPLEVEL(obj);
   AtkObject* child = (AtkObject*)g_list_nth_data(jaw_toplevel->windows, i);
 
-  g_object_ref(G_OBJECT(child));
+  if (G_OBJECT(child) != NULL)
+    g_object_ref(G_OBJECT(child));
 
   return child;
 }
@@ -143,12 +152,14 @@ jaw_toplevel_ref_child (AtkObject *obj, gint i)
 static AtkObject*
 jaw_toplevel_get_parent (AtkObject *obj)
 {
+  JAW_DEBUG_C("%p", obj);
   return NULL;
 }
 
 gint
 jaw_toplevel_add_window (JawToplevel *toplevel, AtkObject *child)
 {
+  JAW_DEBUG_C("%p, %p", toplevel, child);
   if (toplevel == NULL)
   {
     return -1;
@@ -166,6 +177,7 @@ jaw_toplevel_add_window (JawToplevel *toplevel, AtkObject *child)
 
 gint jaw_toplevel_remove_window(JawToplevel *toplevel, AtkObject *child)
 {
+  JAW_DEBUG_C("%p, %p", toplevel, child);
   gint index = -1;
 
   if (toplevel == NULL)
@@ -185,7 +197,7 @@ gint jaw_toplevel_remove_window(JawToplevel *toplevel, AtkObject *child)
 
 gint jaw_toplevel_get_child_index(JawToplevel *toplevel, AtkObject *child)
 {
+  JAW_DEBUG_C("%p, %p", toplevel, child);
   gint i = g_list_index(toplevel->windows, child);
   return i;
 }
-
