@@ -21,68 +21,107 @@ package org.GNOME.Accessibility;
 
 import javax.accessibility.*;
 import javax.swing.*;
+import java.lang.ref.WeakReference;
+
+import java.awt.event.InputEvent;
 import java.awt.event.KeyEvent;
 
 public class AtkAction {
 
-	AccessibleContext ac;
-	AccessibleAction acc_action;
-	AccessibleExtendedComponent acc_ext_component;
+	WeakReference<AccessibleContext> _ac;
+	WeakReference<AccessibleAction> _acc_action;
+	WeakReference<AccessibleExtendedComponent> _acc_ext_component;
+	String[] descriptions;
+	int nactions;
 
 	public AtkAction (AccessibleContext ac) {
 		super();
-		this.ac = ac;
-		this.acc_action = ac.getAccessibleAction();
+		this._ac = new WeakReference<AccessibleContext>(ac);
+		AccessibleAction acc_action = ac.getAccessibleAction();
+		this._acc_action = new WeakReference<AccessibleAction>(acc_action);
+		this.nactions = acc_action.getAccessibleActionCount();
+		this.descriptions = new String[nactions];
 		AccessibleComponent acc_component = ac.getAccessibleComponent();
 		if (acc_component instanceof AccessibleExtendedComponent) {
-			this.acc_ext_component = (AccessibleExtendedComponent)acc_component;
+			this._acc_ext_component = new WeakReference<AccessibleExtendedComponent>((AccessibleExtendedComponent)acc_component);
 		}
 	}
 
-	private class ActionRunner implements Runnable {
-		private AccessibleAction acc_action;
-		private int index;
-
-		public ActionRunner (AccessibleAction acc_action, int index) {
-			this.acc_action = acc_action;
-			this.index = index;
-		}
-
-		public void run () {
-			acc_action.doAccessibleAction(index);
-		}
+	public static AtkAction createAtkAction(AccessibleContext ac){
+		return AtkUtil.invokeInSwing ( () -> { return new AtkAction(ac); }, null);
 	}
-	
+
 	public boolean do_action (int i) {
-		SwingUtilities.invokeLater(new ActionRunner(acc_action, i));
+		AccessibleAction acc_action = _acc_action.get();
+		if (acc_action == null)
+			return false;
+
+		AtkUtil.invokeInSwing( () -> { acc_action.doAccessibleAction(i); });
 		return true;
 	}
 
-	public int get_n_actions () {
-		return acc_action.getAccessibleActionCount();
-	}
+	public int get_n_actions () { return this.nactions; }
 
 	public String get_description (int i) {
-		String description = "<description>";
-		return description;
-	}
+		AccessibleAction acc_action = _acc_action.get();
+		if (acc_action == null)
+			return null;
 
-        public boolean setDescription(int i, String description) {
-               if (description ==  acc_action.getAccessibleActionDescription(i) &&
-                   description != null)
-                     return true;
-               return false;
-        }
-
-	public String get_name (int i) {
-		String name = acc_action.getAccessibleActionDescription(i);
-		if (name == null) {
-			name = " ";
+		if (i >= nactions){
+			return null;
 		}
-
-		return name;
+		if(descriptions[i] != null){
+			return descriptions[i];
+		}
+		descriptions[i] = AtkUtil.invokeInSwing( () -> { return acc_action.getAccessibleActionDescription(i); }, "");
+		return descriptions[i];
 	}
 	
+	public boolean setDescription(int i, String description) {
+		if (i >= nactions){
+			return false;
+		}
+		descriptions[i] = description;
+		return true;
+	}
+
+ /**
+  * @param i an integer holding the index of the name of
+  *          the accessible.
+  * @return  the localized name of the object or otherwise,
+  *          null if the "action" object does not have a
+  *          name (really, java's AccessibleAction class
+  *          does not provide
+  *          a getter for an AccessibleAction
+  *          name so a getter from the AcccessibleContext
+  *          class is one way to work around that)
+  */
+  	public String getLocalizedName (int i) {
+		AccessibleContext ac = _ac.get();
+		if (ac == null)
+			return null;
+		AccessibleAction acc_action = _acc_action.get();
+		if (acc_action == null)
+			return null;
+
+		if (i >= nactions){
+			return null;
+		}
+		if (descriptions[i] != null){
+			return descriptions[i];
+		}
+		return AtkUtil.invokeInSwing ( () -> {
+			descriptions[i] = acc_action.getAccessibleActionDescription(i);
+			if (descriptions[i] != null)
+				return descriptions[i];
+			String name = ac.getAccessibleName();
+			if (name != null)
+				return name;
+			descriptions[i] = "";
+			return descriptions[i];
+		}, null);
+	}
+
 	private String convertModString (String mods) {
 		if (mods == null || mods.length() == 0) {
 			return "";
@@ -93,15 +132,19 @@ public class AtkAction {
 		for (int i = 0; i < modStrs.length; i++) {
 			newModString += "<" + modStrs[i] + ">";
 		}
-		
+
 		return newModString;
 	}
 
 	public String get_keybinding (int index) {
+		AccessibleExtendedComponent acc_ext_component = _acc_ext_component.get();
+		if (_acc_ext_component == null)
+			return "";
+
 		// TODO: improve/fix conversion to strings, concatenate,
 		//       and follow our formatting convention for the role of
 		//       various keybindings (i.e. global, transient, etc.)
-		
+
 		//
 		// Presently, JAA doesn't define a relationship between the index used
 		// and the action associated. As such, all keybindings are only
@@ -110,10 +153,10 @@ public class AtkAction {
 		if (index > 0) {
 			return "";
 		}
-		
+
 		if(acc_ext_component != null) {
 			AccessibleKeyBinding akb = acc_ext_component.getAccessibleKeyBinding();
-			
+
 			if (akb != null && akb.getAccessibleKeyBindingCount() > 0) {
 				String  rVal = "";
 				int     i;
@@ -137,16 +180,16 @@ public class AtkAction {
 				// Since only the first three are relevant, ignore others
 				for (i = 0;( i < akb.getAccessibleKeyBindingCount() && i < 3); i++) {
 					Object o = akb.getAccessibleKeyBinding(i);
-					
+
 					if ( i > 0 ) {
 						rVal += ";";
 					}
-					
-					if (o instanceof javax.swing.KeyStroke) {
-						javax.swing.KeyStroke keyStroke = (javax.swing.KeyStroke)o;
-						String modString = KeyEvent.getKeyModifiersText(keyStroke.getModifiers());
+
+					if (o instanceof KeyStroke) {
+						KeyStroke keyStroke = (KeyStroke)o;
+						String modString = InputEvent.getModifiersExText(keyStroke.getModifiers());
 						String keyString = KeyEvent.getKeyText(keyStroke.getKeyCode());
-						
+
 						if ( keyString != null ) {
 							if ( modString != null && modString.length() > 0 ) {
 								rVal += convertModString(modString) + keyString;
@@ -154,16 +197,16 @@ public class AtkAction {
 								rVal += keyString;
 							}
 						}
-					} else if (o instanceof javax.swing.KeyStroke[]) {
-						javax.swing.KeyStroke[] keyStroke = (javax.swing.KeyStroke[])o;
+					} else if (o instanceof KeyStroke[]) {
+						KeyStroke[] keyStroke = (KeyStroke[])o;
 						for ( int j = 0; j < keyStroke.length; j++ ) {
-							String modString = KeyEvent.getKeyModifiersText(keyStroke[j].getModifiers());
+							String modString = InputEvent.getModifiersExText(keyStroke[j].getModifiers());
 							String keyString = KeyEvent.getKeyText(keyStroke[j].getKeyCode());
-							
+
 							if (j > 0) {
 								rVal += ":";
 							}
-							
+
 							if ( keyString != null ) {
 								if (modString != null && modString.length() > 0) {
 									rVal += convertModString(modString) + keyString;
@@ -174,15 +217,14 @@ public class AtkAction {
 						}
 					}
 				}
-				
+
 				if ( i < 2 ) rVal += ";";
 				if ( i < 3 ) rVal += ";";
-				
+
 				return rVal;
 			}
 		}
-		
+
 		return "";
 	}
 }
-
